#include "ggap/app.h"
#include <moo/singleapp.h>
#include <locale.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <QDebug>
#include <QPlainTextEdit>
#ifdef Q_OS_WIN32
#define _WIN32_WINNT 0x0500
#include <windows.h>
#endif

static bool fatal_warnings = false;
static bool fatal_criticals = false;

#define check_abort(type) \
do { \
    if (type == QtWarningMsg && fatal_warnings) \
        abort(); \
    if (type == QtCriticalMsg && (fatal_warnings || fatal_criticals)) \
        abort(); \
} while (0)

static void win_msg_handler(QtMsgType type, const char *msg)
{
    static QPlainTextEdit *log;

    if (!log)
    {
        log = new QPlainTextEdit;
        log->setAttribute(Qt::WA_QuitOnClose, false);
        log->setAttribute(Qt::WA_DeleteOnClose, false);
    }

    if (!log->isVisible())
        log->show();

    switch (type)
    {
        case QtDebugMsg:
            log->appendPlainText(QString("D: %1").arg(msg));
            break;
        case QtWarningMsg:
            log->appendPlainText(QString("*** Warning: %1").arg(msg));
            break;
        case QtCriticalMsg:
            log->appendPlainText(QString("*** Critical: %1").arg(msg));
            break;
        case QtFatalMsg:
            qInstallMsgHandler(0);
            qFatal("*** Fatal: %s", msg);
            break;
    }
}

static void msg_handler(QtMsgType type, const char *msg)
{
    qInstallMsgHandler(0);

    switch (type)
    {
        case QtDebugMsg:
            qDebug("D: %s", msg);
            break;
        case QtWarningMsg:
            qWarning("*** Warning: %s", msg);
            check_abort(type);
            break;
        case QtCriticalMsg:
            qCritical("*** Critical: %s", msg);
            check_abort(type);
            break;
        case QtFatalMsg:
            qFatal("*** Fatal: %s", msg);
            break;
    }

    qInstallMsgHandler(msg_handler);
}

#ifdef Q_OS_WIN32

#include <windows.h>
#include <stdlib.h>

static void abicheck()
{
    struct S {
        uint a;
        char b;
        uint c : 1;
    };

    // it is 12 with -mms-bitfields
    if (sizeof(S) != 8)
    {
        MessageBoxA(NULL, "Application appears to be incorrectly built, can not continue",
                    NULL, MB_OK | MB_ICONERROR);
        exit(EXIT_FAILURE);
    }
}

#else

static void abicheck()
{
}

#endif

int main(int argc, char *argv[])
{
    abicheck();
    setlocale(LC_ALL, "");

    bool win_handler = false;

    for (int i = 1; i < argc; ++i)
    {
        if (strcmp(argv[i], "--fatal-warnings") == 0)
            fatal_warnings = true;
        else if (strcmp(argv[i], "--fatal-criticals"))
            fatal_criticals = true;
    }

    for (int i = 1; i < argc; )
    {
        if (strcmp(argv[i], "--log-window") == 0)
        {
            win_handler = true;
            memcpy(argv + i, argv + i + 1, sizeof(char*) * (argc - i));
            --argc;
        }
        else if (strcmp(argv[i], "--debug") == 0)
        {
            putenv((char*) "MOO_DEBUG=all");
            memcpy(argv + i, argv + i + 1, sizeof(char*) * (argc - i));
            --argc;
        }
        else
        {
            ++i;
        }
    }

    if (win_handler)
        qInstallMsgHandler(win_msg_handler);
    else
        qInstallMsgHandler(msg_handler);

    ggap::App::parseOptions(argc, argv);

    Q_INIT_RESOURCE(ggap);
    ggap::App app(argc, argv);

    moo::SingleApp sa("ggap.sourceforget.net", argc, argv);
    if (sa.isRunning() && !sa.isDisabled())
    {
#ifdef Q_OS_WIN32
        AllowSetForegroundWindow(sa.processId());
#endif
        QByteArray message = ggap::App::serializeCommandLine();
        if (sa.sendMessage(message, 5000))
            return 0;
    }

    app.connect(&sa, SIGNAL(messageReceived(QByteArray)),
                SLOT(messageReceived(QByteArray)));

    return app.exec();
}
